﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.Xml.Serialization;
using System.IO;
using System.Threading;

namespace BWSceneSwitcher {
    public partial class Form1 :Form {
        public Form1() {
            InitializeComponent();
        }

        public class SceneSwitcherSettings {
            public class HotKeyCombo {
                public bool shift;
                public bool control;
                public bool alt;
                public int hotKeyIndex;

                public HotKeyCombo() {
                    shift = false;
                    control = false;
                    alt = false;
                    hotKeyIndex = 0;
                }
            }

            public HotKeyCombo combo1;
            public HotKeyCombo combo2;
            public HotKeyCombo combo3;
            public HotKeyCombo combo4;
            public HotKeyCombo combo5;
            public HotKeyCombo combo6;
            public bool camtasiaHotkeys;

            public SceneSwitcherSettings() {
                combo1 = new HotKeyCombo();
                combo2 = new HotKeyCombo();
                combo3 = new HotKeyCombo();
                combo4 = new HotKeyCombo();
                combo5 = new HotKeyCombo();
                combo6 = new HotKeyCombo();
                camtasiaHotkeys = false;
            }
        }


        private MemoryAccess.GameState lastGameState = MemoryAccess.GameState.InGame;
        private Dictionary<String, VirtualKeyCode> hotKeyList = new Dictionary<String, VirtualKeyCode>();
        private string settingsPath = Environment.GetFolderPath(Environment.SpecialFolder.LocalApplicationData);

        private void timer1_Tick(object sender, EventArgs e) {
            // Check if BroodWar is running
            if (MemoryAccess.GetBroodWarHWND() == IntPtr.Zero) {
                MemoryAccess.CloseBroodWarHandle();
            } else {
                MemoryAccess.OpenBroodwarProc();
            }

            VirtualKeyCode keyCode;
            MemoryAccess.GameState gameState = MemoryAccess.GetGameState();
            // check if game state changed and it's not an unknow game state and its not switching from ingame to pregame lobby (as thats not possible ;-))
            if (lastGameState == gameState || 
                gameState == MemoryAccess.GameState.Other || 
                (lastGameState == MemoryAccess.GameState.InGame && gameState == MemoryAccess.GameState.PreGame_Lobby)) 
            {
                return;
            }
            lastGameState = gameState;
            // stop and restart camtasia recording
            if (chkCamtasiaHK.Checked) {
                hotKeyList.TryGetValue(cmbHotkeys5.SelectedItem.ToString(), out keyCode);
                SimulteHotkeyCombo(chkShift5.Checked, chkControl5.Checked, chkAlt5.Checked, keyCode);
                System.Threading.Thread.Sleep(200);
                hotKeyList.TryGetValue(cmbHotkeys6.SelectedItem.ToString(), out keyCode);
                SimulteHotkeyCombo(chkShift6.Checked, chkControl6.Checked, chkAlt6.Checked, keyCode);
            }
            
            switch (gameState) {
                case MemoryAccess.GameState.Battle_Net_Menu:
                    if (cmbHotkeys1.SelectedItem.ToString() == "None") {
                        break;
                    }
                    hotKeyList.TryGetValue(cmbHotkeys1.SelectedItem.ToString(), out keyCode);
                    SimulteHotkeyCombo(chkShift1.Checked, chkControl1.Checked, chkAlt1.Checked, keyCode);
                    Log("Switched to Battle.Net Menu Scene");
                    break;
                case MemoryAccess.GameState.PreGame_Lobby:
                    if (cmbHotkeys2.SelectedItem.ToString() == "None") {
                        break;
                    }
                    hotKeyList.TryGetValue(cmbHotkeys2.SelectedItem.ToString(), out keyCode);
                    SimulteHotkeyCombo(chkShift2.Checked, chkControl2.Checked, chkAlt2.Checked, keyCode);
                    Log("Switched to PreGame Lobby Scene");
                    break;
                case MemoryAccess.GameState.InGame:
                    if (cmbHotkeys3.SelectedItem.ToString() == "None") {
                        break;
                    }
                    hotKeyList.TryGetValue(cmbHotkeys3.SelectedItem.ToString(), out keyCode);
                    SimulteHotkeyCombo(chkShift3.Checked, chkControl3.Checked, chkAlt3.Checked, keyCode);
                    Log("Switched to InGame Scene");
                    break;
                case MemoryAccess.GameState.Minimized:
                    if (cmbHotkeys4.SelectedItem.ToString() == "None") {
                        break;
                    }
                    hotKeyList.TryGetValue(cmbHotkeys4.SelectedItem.ToString(), out keyCode);
                    SimulteHotkeyCombo(chkShift4.Checked, chkControl4.Checked, chkAlt4.Checked, keyCode);
                    Log("Switched to Minimized Scene");
                    break;
            }
        }

        private void Log(string text) {
            this.richTextBox1.Text += DateTime.Now.ToLongTimeString() + " " + text + Environment.NewLine;
        }

        private void SimulteHotkeyCombo(bool shift, bool control, bool alt, VirtualKeyCode keyCode) {
            // OBS has some weird-ass conditional check that misuses GetAsyncKeyState, so it needs
            // to recognize that keys are down in one frame and up in a following frame to realize
            // that they've been pressed. Its scanning rate is 30ms, so we need to go at least
            // double that so we ensure we're hitting successive frames
            const int KEY_DELAY = 70;
            if (shift) {
                InputSimulator.SimulateKeyDown(VirtualKeyCode.SHIFT);
                Thread.Sleep(KEY_DELAY);
            }
            if (control) {
                InputSimulator.SimulateKeyDown(VirtualKeyCode.CONTROL);
                Thread.Sleep(KEY_DELAY);
            }
            if (alt) {
                InputSimulator.SimulateKeyDown(VirtualKeyCode.MENU);
                Thread.Sleep(KEY_DELAY);
            }
            InputSimulator.SimulateKeyDown(keyCode);
            Thread.Sleep(KEY_DELAY);
            InputSimulator.SimulateKeyUp(keyCode);
            if (alt) {
                Thread.Sleep(KEY_DELAY);
                InputSimulator.SimulateKeyUp(VirtualKeyCode.MENU);
            }
            if (control) {
                Thread.Sleep(KEY_DELAY);
                InputSimulator.SimulateKeyUp(VirtualKeyCode.CONTROL);
            }
            if (shift) {
                Thread.Sleep(KEY_DELAY);
                InputSimulator.SimulateKeyUp(VirtualKeyCode.SHIFT);
            }
        }

        private void Form1_Load(object sender, EventArgs e) {
            FillComboBoxes();
            LoadSettings();

            // Get the rights to read BroodWars memory
            MemoryAccess.SetDebugPrivileges();
            MemoryAccess.OpenBroodwarProc();

            // Start the timer which checks for scene switches
            timer1.Start();
        }

        private void SaveSettings(SceneSwitcherSettings settings) {
            XmlSerializer serializer = new XmlSerializer(typeof(SceneSwitcherSettings));
            TextWriter textWriter = new StreamWriter(Path.Combine(settingsPath, "settings.xml"));
            serializer.Serialize(textWriter, settings);
            textWriter.Close();
        }

        private void FillComboBoxes() {
            hotKeyList.Add("None", VirtualKeyCode.NONAME);

            for (int i=1; i<=12; i++) {
                VirtualKeyCode vkc = (VirtualKeyCode)((int)(VirtualKeyCode.F1) + (i-1));
                hotKeyList.Add("F" + i.ToString(), vkc);
            }

            for (int i=0; i<=9; i++) {
                VirtualKeyCode vkc = (VirtualKeyCode)((int)(VirtualKeyCode.NUMPAD0) + i);
                hotKeyList.Add("NumPad" + i.ToString(), vkc);
            }

            for (int i=0; i<=9; i++) {
                VirtualKeyCode vkc = (VirtualKeyCode)((int)(VirtualKeyCode.VK_0) + i);
                hotKeyList.Add(i.ToString(), vkc);
            }

            for (char c='A'; c <= 'Z'; c++) {
                VirtualKeyCode vkc = (VirtualKeyCode)((int)(VirtualKeyCode.VK_A) + (int)(c - 'A'));
                hotKeyList.Add(c.ToString(), vkc);
            }

            foreach (string s in hotKeyList.Keys) {
                cmbHotkeys1.Items.Add(s);
                cmbHotkeys1.SelectedIndex = 0;
                cmbHotkeys2.Items.Add(s);
                cmbHotkeys2.SelectedIndex = 0;
                cmbHotkeys3.Items.Add(s);
                cmbHotkeys3.SelectedIndex = 0;
                cmbHotkeys4.Items.Add(s);
                cmbHotkeys4.SelectedIndex = 0;
                cmbHotkeys5.Items.Add(s);
                cmbHotkeys5.SelectedIndex = 0;
                cmbHotkeys6.Items.Add(s);
                cmbHotkeys6.SelectedIndex = 0;
            }
        }

        private void Form1_FormClosing(object sender, FormClosingEventArgs e) {
            SceneSwitcherSettings settings = new SceneSwitcherSettings();
            settings.combo1.shift = chkShift1.Checked;
            settings.combo2.shift = chkShift2.Checked;
            settings.combo3.shift = chkShift3.Checked;
            settings.combo4.shift = chkShift4.Checked;
            settings.combo5.shift = chkShift5.Checked;
            settings.combo6.shift = chkShift6.Checked;

            settings.combo1.control = chkControl1.Checked;
            settings.combo2.control = chkControl2.Checked;
            settings.combo3.control = chkControl3.Checked;
            settings.combo4.control = chkControl4.Checked;
            settings.combo5.control = chkControl5.Checked;
            settings.combo6.control = chkControl6.Checked;

            settings.combo1.alt = chkAlt1.Checked;
            settings.combo2.alt = chkAlt2.Checked;
            settings.combo3.alt = chkAlt3.Checked;
            settings.combo4.alt = chkAlt4.Checked;
            settings.combo5.alt = chkAlt5.Checked;
            settings.combo6.alt = chkAlt6.Checked;

            settings.combo1.hotKeyIndex = cmbHotkeys1.SelectedIndex;
            settings.combo2.hotKeyIndex = cmbHotkeys2.SelectedIndex;
            settings.combo3.hotKeyIndex = cmbHotkeys3.SelectedIndex;
            settings.combo4.hotKeyIndex = cmbHotkeys4.SelectedIndex;
            settings.combo5.hotKeyIndex = cmbHotkeys5.SelectedIndex;
            settings.combo6.hotKeyIndex = cmbHotkeys6.SelectedIndex;

            settings.camtasiaHotkeys = chkCamtasiaHK.Checked;

            SaveSettings(settings);
        }

        private void LoadSettings() {
            XmlSerializer deserializer = new XmlSerializer(typeof(SceneSwitcherSettings));
            TextReader textReader = new StreamReader(Path.Combine(settingsPath, "settings.xml"));
            SceneSwitcherSettings settings = new SceneSwitcherSettings();
            settings = (SceneSwitcherSettings)deserializer.Deserialize(textReader);
            textReader.Close();

            chkShift1.Checked = settings.combo1.shift;
            chkShift2.Checked = settings.combo2.shift;
            chkShift3.Checked = settings.combo3.shift;
            chkShift4.Checked = settings.combo4.shift;
            chkShift5.Checked = settings.combo5.shift;
            chkShift6.Checked = settings.combo6.shift;

            chkControl1.Checked = settings.combo1.control;
            chkControl2.Checked = settings.combo2.control;
            chkControl3.Checked = settings.combo3.control;
            chkControl4.Checked = settings.combo4.control;
            chkControl5.Checked = settings.combo5.control;
            chkControl6.Checked = settings.combo6.control;

            chkAlt1.Checked = settings.combo1.alt;
            chkAlt2.Checked = settings.combo2.alt;
            chkAlt3.Checked = settings.combo3.alt;
            chkAlt4.Checked = settings.combo4.alt;
            chkAlt5.Checked = settings.combo5.alt;
            chkAlt6.Checked = settings.combo6.alt;

            cmbHotkeys1.SelectedIndex = settings.combo1.hotKeyIndex;
            cmbHotkeys2.SelectedIndex = settings.combo2.hotKeyIndex;
            cmbHotkeys3.SelectedIndex = settings.combo3.hotKeyIndex;
            cmbHotkeys4.SelectedIndex = settings.combo4.hotKeyIndex;
            cmbHotkeys5.SelectedIndex = settings.combo5.hotKeyIndex;
            cmbHotkeys6.SelectedIndex = settings.combo6.hotKeyIndex;

            chkCamtasiaHK.Checked = settings.camtasiaHotkeys;
        }
    }
}
