﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Runtime.InteropServices;

namespace BWSceneSwitcher {
    [StructLayout(LayoutKind.Sequential)]
    internal struct HARDWAREINPUT {
        public uint Msg;
        public ushort ParamL;
        public ushort ParamH;
    }

    [StructLayout(LayoutKind.Sequential)]
    internal struct INPUT {
        public uint Type;
        public MOUSEKEYBDHARDWAREINPUT Data;
    }

    public enum InputType :uint {
        HARDWARE = 2,
        KEYBOARD = 1,
        MOUSE = 0
    }

    [StructLayout(LayoutKind.Sequential)]
    internal struct KEYBDINPUT {
        public ushort Vk;
        public ushort Scan;
        public uint Flags;
        public uint Time;
        public IntPtr ExtraInfo;
    }

    public enum KeyboardFlag :uint {
        EXTENDEDKEY = 1,
        KEYUP = 2,
        SCANCODE = 8,
        UNICODE = 4
    }

    public enum MouseFlag :uint {
        ABSOLUTE = 0x8000,
        LEFTDOWN = 2,
        LEFTUP = 4,
        MIDDLEDOWN = 0x20,
        MIDDLEUP = 0x40,
        MOVE = 1,
        RIGHTDOWN = 8,
        RIGHTUP = 0x10,
        VIRTUALDESK = 0x4000,
        WHEEL = 0x800,
        XDOWN = 0x80,
        XUP = 0x100
    }

    [StructLayout(LayoutKind.Sequential)]
    internal struct MOUSEINPUT {
        public int X;
        public int Y;
        public uint MouseData;
        public uint Flags;
        public uint Time;
        public IntPtr ExtraInfo;
    }

    [StructLayout(LayoutKind.Explicit)]
    internal struct MOUSEKEYBDHARDWAREINPUT {
        // Fields
        [FieldOffset(0)]
        public HARDWAREINPUT Hardware;
        [FieldOffset(0)]
        public KEYBDINPUT Keyboard;
        [FieldOffset(0)]
        public MOUSEINPUT Mouse;
    }

    public enum VirtualKeyCode :ushort {
        ACCEPT = 30,
        ADD = 0x6b,
        APPS = 0x5d,
        ATTN = 0xf6,
        BACK = 8,
        BROWSER_BACK = 0xa6,
        BROWSER_FAVORITES = 0xab,
        BROWSER_FORWARD = 0xa7,
        BROWSER_HOME = 0xac,
        BROWSER_REFRESH = 0xa8,
        BROWSER_SEARCH = 170,
        BROWSER_STOP = 0xa9,
        CANCEL = 3,
        CAPITAL = 20,
        CLEAR = 12,
        CONTROL = 0x11,
        CONVERT = 0x1c,
        CRSEL = 0xf7,
        DECIMAL = 110,
        DELETE = 0x2e,
        DIVIDE = 0x6f,
        DOWN = 40,
        END = 0x23,
        EREOF = 0xf9,
        ESCAPE = 0x1b,
        EXECUTE = 0x2b,
        EXSEL = 0xf8,
        F1 = 0x70,
        F10 = 0x79,
        F11 = 0x7a,
        F12 = 0x7b,
        F13 = 0x7c,
        F14 = 0x7d,
        F15 = 0x7e,
        F16 = 0x7f,
        F17 = 0x80,
        F18 = 0x81,
        F19 = 130,
        F2 = 0x71,
        F20 = 0x83,
        F21 = 0x84,
        F22 = 0x85,
        F23 = 0x86,
        F24 = 0x87,
        F3 = 0x72,
        F4 = 0x73,
        F5 = 0x74,
        F6 = 0x75,
        F7 = 0x76,
        F8 = 0x77,
        F9 = 120,
        FINAL = 0x18,
        HANGEUL = 0x15,
        HANGUL = 0x15,
        HANJA = 0x19,
        HELP = 0x2f,
        HOME = 0x24,
        INSERT = 0x2d,
        JUNJA = 0x17,
        KANA = 0x15,
        KANJI = 0x19,
        LAUNCH_APP1 = 0xb6,
        LAUNCH_APP2 = 0xb7,
        LAUNCH_MAIL = 180,
        LAUNCH_MEDIA_SELECT = 0xb5,
        LBUTTON = 1,
        LCONTROL = 0xa2,
        LEFT = 0x25,
        LMENU = 0xa4,
        LSHIFT = 160,
        LWIN = 0x5b,
        MBUTTON = 4,
        MEDIA_NEXT_TRACK = 0xb0,
        MEDIA_PLAY_PAUSE = 0xb3,
        MEDIA_PREV_TRACK = 0xb1,
        MEDIA_STOP = 0xb2,
        MENU = 0x12,
        MODECHANGE = 0x1f,
        MULTIPLY = 0x6a,
        NEXT = 0x22,
        NONAME = 0xfc,
        NONCONVERT = 0x1d,
        NUMLOCK = 0x90,
        NUMPAD0 = 0x60,
        NUMPAD1 = 0x61,
        NUMPAD2 = 0x62,
        NUMPAD3 = 0x63,
        NUMPAD4 = 100,
        NUMPAD5 = 0x65,
        NUMPAD6 = 0x66,
        NUMPAD7 = 0x67,
        NUMPAD8 = 0x68,
        NUMPAD9 = 0x69,
        OEM_1 = 0xba,
        OEM_102 = 0xe2,
        OEM_2 = 0xbf,
        OEM_3 = 0xc0,
        OEM_4 = 0xdb,
        OEM_5 = 220,
        OEM_6 = 0xdd,
        OEM_7 = 0xde,
        OEM_8 = 0xdf,
        OEM_CLEAR = 0xfe,
        OEM_COMMA = 0xbc,
        OEM_MINUS = 0xbd,
        OEM_PERIOD = 190,
        OEM_PLUS = 0xbb,
        PA1 = 0xfd,
        PACKET = 0xe7,
        PAUSE = 0x13,
        PLAY = 250,
        PRINT = 0x2a,
        PRIOR = 0x21,
        PROCESSKEY = 0xe5,
        RBUTTON = 2,
        RCONTROL = 0xa3,
        RETURN = 13,
        RIGHT = 0x27,
        RMENU = 0xa5,
        RSHIFT = 0xa1,
        RWIN = 0x5c,
        SCROLL = 0x91,
        SELECT = 0x29,
        SEPARATOR = 0x6c,
        SHIFT = 0x10,
        SLEEP = 0x5f,
        SNAPSHOT = 0x2c,
        SPACE = 0x20,
        SUBTRACT = 0x6d,
        TAB = 9,
        UP = 0x26,
        VK_0 = 0x30,
        VK_1 = 0x31,
        VK_2 = 50,
        VK_3 = 0x33,
        VK_4 = 0x34,
        VK_5 = 0x35,
        VK_6 = 0x36,
        VK_7 = 0x37,
        VK_8 = 0x38,
        VK_9 = 0x39,
        VK_A = 0x41,
        VK_B = 0x42,
        VK_C = 0x43,
        VK_D = 0x44,
        VK_E = 0x45,
        VK_F = 70,
        VK_G = 0x47,
        VK_H = 0x48,
        VK_I = 0x49,
        VK_J = 0x4a,
        VK_K = 0x4b,
        VK_L = 0x4c,
        VK_M = 0x4d,
        VK_N = 0x4e,
        VK_O = 0x4f,
        VK_P = 80,
        VK_Q = 0x51,
        VK_R = 0x52,
        VK_S = 0x53,
        VK_T = 0x54,
        VK_U = 0x55,
        VK_V = 0x56,
        VK_W = 0x57,
        VK_X = 0x58,
        VK_Y = 0x59,
        VK_Z = 90,
        VOLUME_DOWN = 0xae,
        VOLUME_MUTE = 0xad,
        VOLUME_UP = 0xaf,
        XBUTTON1 = 5,
        XBUTTON2 = 6,
        ZOOM = 0xfb
    }

    public enum XButton :uint {
        XBUTTON1 = 1,
        XBUTTON2 = 2
    }


    public static class InputSimulator {
        // Methods
        [DllImport("user32.dll", SetLastError=true)]
        private static extern short GetAsyncKeyState(ushort virtualKeyCode);
        [DllImport("user32.dll", SetLastError=true)]
        private static extern short GetKeyState(ushort virtualKeyCode);
        [DllImport("user32.dll")]
        private static extern IntPtr GetMessageExtraInfo();
        public static bool IsKeyDown(VirtualKeyCode keyCode) {
            return (GetKeyState((ushort)keyCode) < 0);
        }

        public static bool IsKeyDownAsync(VirtualKeyCode keyCode) {
            return (GetAsyncKeyState((ushort)keyCode) < 0);
        }

        public static bool IsTogglingKeyInEffect(VirtualKeyCode keyCode) {
            return ((GetKeyState((ushort)keyCode) & 1) == 1);
        }

        [DllImport("user32.dll", SetLastError=true)]
        private static extern uint SendInput(uint numberOfInputs, INPUT[] inputs, int sizeOfInputStructure);
        public static void SimulateKeyDown(VirtualKeyCode keyCode) {
            INPUT input = new INPUT();
            input.Type = 1;
            input.Data.Keyboard = new KEYBDINPUT();
            input.Data.Keyboard.Vk = (ushort)keyCode;
            input.Data.Keyboard.Scan = 0;
            input.Data.Keyboard.Flags = 0;
            input.Data.Keyboard.Time = 0;
            input.Data.Keyboard.ExtraInfo = IntPtr.Zero;
            INPUT[] inputs = new INPUT[] { input };
            if (SendInput(1, inputs, Marshal.SizeOf(typeof(INPUT))) == 0) {
                throw new Exception(string.Format("The key down simulation for {0} was not successful.", keyCode));
            }
        }

        public static void SimulateKeyPress(VirtualKeyCode keyCode) {
            INPUT input = new INPUT();
            input.Type = 1;
            input.Data.Keyboard = new KEYBDINPUT();
            input.Data.Keyboard.Vk = (ushort)keyCode;
            input.Data.Keyboard.Scan = 0;
            input.Data.Keyboard.Flags = 0;
            input.Data.Keyboard.Time = 0;
            input.Data.Keyboard.ExtraInfo = IntPtr.Zero;
            INPUT input2 = new INPUT();
            input2.Type = 1;
            input2.Data.Keyboard = new KEYBDINPUT();
            input2.Data.Keyboard.Vk = (ushort)keyCode;
            input2.Data.Keyboard.Scan = 0;
            input2.Data.Keyboard.Flags = 2;
            input2.Data.Keyboard.Time = 0;
            input2.Data.Keyboard.ExtraInfo = IntPtr.Zero;
            INPUT[] inputs = new INPUT[] { input, input2 };
            if (SendInput(2, inputs, Marshal.SizeOf(typeof(INPUT))) == 0) {
                throw new Exception(string.Format("The key press simulation for {0} was not successful.", keyCode));
            }
        }

        public static void SimulateKeyUp(VirtualKeyCode keyCode) {
            INPUT input = new INPUT();
            input.Type = 1;
            input.Data.Keyboard = new KEYBDINPUT();
            input.Data.Keyboard.Vk = (ushort)keyCode;
            input.Data.Keyboard.Scan = 0;
            input.Data.Keyboard.Flags = 2;
            input.Data.Keyboard.Time = 0;
            input.Data.Keyboard.ExtraInfo = IntPtr.Zero;
            INPUT[] inputs = new INPUT[] { input };
            if (SendInput(1, inputs, Marshal.SizeOf(typeof(INPUT))) == 0) {
                throw new Exception(string.Format("The key up simulation for {0} was not successful.", keyCode));
            }
        }

        public static void SimulateTextEntry(string text) {
            if (text.Length > 0x7fffffffL) {
                throw new ArgumentException(string.Format("The text parameter is too long. It must be less than {0} characters.", 0x7fffffff), "text");
            }
            byte[] bytes = Encoding.ASCII.GetBytes(text);
            int length = bytes.Length;
            INPUT[] inputs = new INPUT[length * 2];
            for (int i = 0; i < length; i++) {
                ushort num3 = bytes[i];
                INPUT input = new INPUT();
                input.Type = 1;
                input.Data.Keyboard = new KEYBDINPUT();
                input.Data.Keyboard.Vk = 0;
                input.Data.Keyboard.Scan = num3;
                input.Data.Keyboard.Flags = 4;
                input.Data.Keyboard.Time = 0;
                input.Data.Keyboard.ExtraInfo = IntPtr.Zero;
                INPUT input2 = new INPUT();
                input2.Type = 1;
                input2.Data.Keyboard = new KEYBDINPUT();
                input2.Data.Keyboard.Vk = 0;
                input2.Data.Keyboard.Scan = num3;
                input2.Data.Keyboard.Flags = 6;
                input2.Data.Keyboard.Time = 0;
                input2.Data.Keyboard.ExtraInfo = IntPtr.Zero;
                if ((num3 & 0xff00) == 0xe000) {
                    input.Data.Keyboard.Flags |= 1;
                    input2.Data.Keyboard.Flags |= 1;
                }
                inputs[2 * i] = input;
                inputs[(2 * i) + 1] = input2;
            }
            SendInput((uint)(length * 2), inputs, Marshal.SizeOf(typeof(INPUT)));
        }
    }
}
